/* Comparator.java
 *
 * created: Mon Jul 12 1999,2000
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1999,2000,2001  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/pathsoft/repository/powmap/diana/components/Comparator.java,v 1.26 2001/09/13 09:22:43 kmr Exp $
 */

package diana.components;

import AppGlobal;

import diana.*;
import diana.plot.*;
import diana.sequence.*;

import uk.ac.sanger.pathogens.OutOfRangeException;
import uk.ac.sanger.pathogens.embl.EntryInformationException;
import uk.ac.sanger.pathogens.embl.Range;
import uk.ac.sanger.pathogens.embl.DocumentEntryFactory;
import uk.ac.sanger.pathogens.ReadOnlyException;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Vector;
import java.util.EventObject;

/**
 *  This Frame component contains a AlignmentViewer component, two
 *  FeatureDisplay components and the event listeners that keep them
 *  synchronized.
 *
 *  @author Kim Rutherford
 *  @version $Id: Comparator.java,v 1.26 2001/09/13 09:22:43 kmr Exp $
 **/

public class Comparator extends Frame
    implements FeatureDisplayOwner {
  /**
   *  Create a new Comparator component using the given data.  The component
   *  will contain two FeatureDisplay components and one AlignmentViewer
   *  components.
   *  @param comparison_data Provides the AlignMatch objects that will be
   *    displayed.
   *  @param subject_entry_group The entry_group to show at the top
   *  @param query_entry_group The entry_group to show at the bottom
   **/
  public Comparator (final EntryGroup subject_entry_group,
                     final EntryGroup query_entry_group,
                     final ComparisonData comparison_data) {
    this.subject_entry_group = subject_entry_group;
    this.query_entry_group = query_entry_group;
    this.comparison_data = comparison_data;

    // XXX add a InputStreamProgressListener
    this.entry_sources = AppGlobal.getEntrySources (this, null);

    button_panel = new Panel ();

    final FlowLayout flow_layout = new FlowLayout (FlowLayout.LEFT);

    button_panel.setLayout (flow_layout);

    makeSelections ();
    makeGotoEventSources ();

    subject_feature_display =
      new FeatureDisplay (getSubjectEntryGroup (), getSubjectSelection (),
                          getSubjectGotoEventSource (), this, true);

    query_feature_display =
      new FeatureDisplay (getQueryEntryGroup (), getQuerySelection (),
                          getQueryGotoEventSource (), this, false);

    alignment_viewer =
      new AlignmentViewer (getSubjectDisplay () ,
                           getQueryDisplay (),
                           comparison_data);

    final ComparatorGlue comparator_glue =
      new ComparatorGlue (this,
                          getSubjectDisplay (), getQueryDisplay (),
                          getAlignmentViewer ());

    subject_feature_display.setShowLabels (false);
    query_feature_display.setShowLabels (false);

    setLayout(new BorderLayout ());

    setTitle ("ACT");

    if (getSubjectEntryGroup ().getDefaultEntry () != null &&
        getQueryEntryGroup ().getDefaultEntry () != null) {
      final String first_name =
        getSubjectEntryGroup ().getDefaultEntry ().getName ();
      final String second_name =
        getQueryEntryGroup ().getDefaultEntry ().getName ();

      if (first_name != null && second_name != null) {
        setTitle ("ACT: " + first_name + " vs " +
                  second_name);
      }
    }

    subject_plots =
      new BasePlotGroup (getSubjectEntryGroup (), this,
                         getSubjectSelection (), getSubjectGotoEventSource ());
    query_plots =
      new BasePlotGroup (getQueryEntryGroup (), this, getQuerySelection (),
                         getQueryGotoEventSource ());

    subject_feature_display.addDisplayAdjustmentListener (subject_plots);
    query_feature_display.addDisplayAdjustmentListener (query_plots);

    final EntryGroupChangeListener sub_change_listener =
      new EntryGroupChangeListener () {
        public void entryGroupChanged (final EntryGroupChangeEvent event) {
          switch (event.getType ()) {
          case EntryGroupChangeEvent.ENTRY_ADDED:
          case EntryGroupChangeEvent.ENTRY_DELETED:
            makeFileMenu ();
            break;
          }
        }
      };

    final EntryGroupChangeListener query_change_listener =
      new EntryGroupChangeListener () {
        public void entryGroupChanged (final EntryGroupChangeEvent event) {
          switch (event.getType ()) {
          case EntryGroupChangeEvent.ENTRY_ADDED:
          case EntryGroupChangeEvent.ENTRY_DELETED:
            makeFileMenu ();
            break;
          }
        }
      };

    getSubjectEntryGroup ().addEntryGroupChangeListener (sub_change_listener);
    getQueryEntryGroup ().addEntryGroupChangeListener (query_change_listener);

    final Font default_font = getDefaultFont ();

    setFont (default_font);

    makeMenus ();

    GridBagLayout gridbag = new GridBagLayout();
    setLayout (gridbag);

    GridBagConstraints c = new GridBagConstraints();

    c.gridwidth  = GridBagConstraints.REMAINDER;
    c.fill       = GridBagConstraints.HORIZONTAL;
    c.anchor     = GridBagConstraints.NORTH;
    c.gridheight = 1;
    c.weightx    = 1;
    c.weighty    = 0;

    gridbag.setConstraints (button_panel, c);
    add (button_panel);

    c.weighty = 0;
    gridbag.setConstraints (subject_plots, c);
    add (subject_plots);

    gridbag.setConstraints (subject_feature_display, c);
    add (subject_feature_display);

    c.fill = GridBagConstraints.BOTH;
    c.weighty = 1;
    gridbag.setConstraints (getAlignmentViewer (), c);
    add (alignment_viewer);

    c.weighty = 0;
    gridbag.setConstraints (query_feature_display, c);
    add (query_feature_display);

    c.weighty = 0;
    gridbag.setConstraints (query_plots, c);
    add (query_plots);

    addWindowListener (new WindowAdapter () {
      public void windowClosing (WindowEvent event) {
        Comparator.this.dispose ();
      }
    });

    packme ();

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();

    int screen_height = screen.height;
    int screen_width = screen.width;

    if (screen_width <= 900 || screen_height <= 700) {
      setSize (screen_width * 9 / 10, screen_height * 9 / 10);
    } else {
      setSize (900, 700);
    }

    setLocation (new Point ((screen.width - getSize ().width) / 2,
                            (screen.height - getSize ().height) / 2));
  }

  /**
   *  This method will call pack () and then move the Frame to the centre of
   *  the screen.  (Implementation of the FeatureDisplayOwner interface).
   **/
  public void packme () {
    pack ();

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();

    final int x_position = (screen.width - getSize ().width) / 2;
    int y_position = (screen.height - getSize ().height) / 2;

    if (y_position < 10) {
      y_position = 10;
    }

    setLocation (new Point (x_position, y_position));
  }

  /**
   *  Return the reference of the subject FeatureDisplay.
   **/
  public FeatureDisplay getSubjectDisplay () {
    return subject_feature_display;
  }

  /**
   *  Return the reference of the query FeatureDisplay.
   **/
  public FeatureDisplay getQueryDisplay () {
    return query_feature_display;
  }

  /**
   *  Return the reference of the EntryGroup in view in the subject
   *  FeatureDisplay.
   **/
  public EntryGroup getSubjectEntryGroup () {
    return subject_entry_group;
  }

  /**
   *  Return the reference of the EntryGroup in view in the query
   *  FeatureDisplay.
   **/
  public EntryGroup getQueryEntryGroup () {
    return query_entry_group;
  }
 
  /**
   *  Return the Selection object used by the subject FeatureDisplay.
   **/
  public Selection getSubjectSelection () {
    return subject_selection;
  }

  /**
   *  Return the Selection object used by the query FeatureDisplay.
   **/
  public Selection getQuerySelection () {
    return query_selection;
  }

  /**
   *  Return the AlignmentViewer that was created in the constructor.
   **/
  private AlignmentViewer getAlignmentViewer () {
    return alignment_viewer;
  }

  /**
   *  Return the GotoEventSource object used by the subject FeatureDisplay.
   **/
  public GotoEventSource getSubjectGotoEventSource () {
    return subject_goto_event_source;
  }

  /**
   *  Return the GotoEventSource object used by the query FeatureDisplay.
   **/
  public GotoEventSource getQueryGotoEventSource () {
    return query_goto_event_source;
  }

  /**
   *  Make subject_goto_event_source and query_goto_event_source and wire them
   *  togeather so that goto events do the right thing.
   **/
  private void makeGotoEventSources () {
    subject_goto_event_source =
      new SimpleGotoEventSource (getSubjectEntryGroup ()) {
        /**
         *  Send the given event to all the GotoListeners.
         **/
        public void sendGotoEvent (final GotoEvent goto_event) {
          // temporarily disable moving so that the other FeatureDisplay
          // doesn't start moving (because of the listeners set up in
          // addDisplayListeners ()
          super.sendGotoEvent (goto_event);
        }
      };

    query_goto_event_source =
      new SimpleGotoEventSource (getQueryEntryGroup ()) {
        /**
         *  Send the given event to all the GotoListeners.
         **/
        public void sendGotoEvent (final GotoEvent goto_event) {
          // temporarily disable moving so that the other FeatureDisplay
          // doesn't start moving (because of the listeners set up in
          // addDisplayListeners ()
          super.sendGotoEvent (goto_event);
        }
      };
  }

  /**
   *  Make a Selection object for the subject and one for the query.  Add
   *  SelectionChangeListeners for each that update the selected matches in
   *  the AlignmentViewer.
   **/
  private void makeSelections () {
    subject_selection = new Selection (null);
    query_selection = new Selection (null);
  }

  /**
   *  Return the current default font (from Diana.options).
   **/
  private Font getDefaultFont () {
    return Options.getOptions ().getFont ();
  }

  /**
   *  Make and add the menus for this component.
   **/
  private void makeMenus () {
    final Font default_font = getDefaultFont ();

    setMenuBar (menu_bar);

    makeFileMenu ();

    menu_bar.add (file_menu);

    final BaseAlgorithm [] subject_algorithms =
      subject_plots.getPlotAlgorithms ();

    for (int i = 0 ; i < subject_algorithms.length ; ++i) {
      final BaseAlgorithm this_algorithm = subject_algorithms[i];

      addDisplayMenuAlgorithm (subject_plots, this_algorithm,
                               subject_graphs_menu);
    }

    final BaseAlgorithm [] query_algorithms =
      query_plots.getPlotAlgorithms ();

    for (int i = 0 ; i < query_algorithms.length ; ++i) {
      final BaseAlgorithm this_algorithm = query_algorithms[i];

      addDisplayMenuAlgorithm (query_plots, this_algorithm,
                               query_graphs_menu);
    }

    menu_bar.add (subject_graphs_menu);
    menu_bar.add (query_graphs_menu);

    final Menu display_menu = new Menu ("Display");


    final MenuItem hide_on_frame_lines_item =
      new MenuItem ("Hide Frame Lines");
    hide_on_frame_lines_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        getSubjectDisplay ().setShowForwardFrameLines (false);
        getSubjectDisplay ().setShowReverseFrameLines (false);
        getQueryDisplay ().setShowForwardFrameLines (false);
        getQueryDisplay ().setShowReverseFrameLines (false);
      }
    });

    display_menu.add (hide_on_frame_lines_item);


    final MenuItem show_on_frame_lines_item =
      new MenuItem ("Show Frame Lines");
    show_on_frame_lines_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        getSubjectDisplay ().setShowForwardFrameLines (true);
        getSubjectDisplay ().setShowReverseFrameLines (true);
        getQueryDisplay ().setShowForwardFrameLines (true);
        getQueryDisplay ().setShowReverseFrameLines (true);
      }
    });

    display_menu.add (show_on_frame_lines_item);


    menu_bar.add (display_menu);
  }

  /**
   *  Add a menu item for the given algorithm to the given menu.
   **/
  private void addDisplayMenuAlgorithm (final BasePlotGroup base_plot_group,
                                        final BaseAlgorithm algorithm,
                                        final Menu menu) {
    final CheckboxMenuItem new_item =
      new CheckboxMenuItem (algorithm.getAlgorithmName ());

    new_item.setState (base_plot_group.basePlotIsVisible (algorithm));

    new_item.addItemListener (new ItemListener () {
      public void itemStateChanged(ItemEvent event) {
        base_plot_group.setVisibleByAlgorithm (algorithm,
                                               new_item.getState ());
        packme ();
      }
    });

    menu.add (new_item);
  }

  /**
   *  Save the given entry, prompting for a file name if necessary.
   **/
  private void saveEntry (final Entry entry) {
    try {
      if (entry.getName () == null) {
        // prompt for a name
        final FileDialog dialog =
          new FileDialog (this,
                          "Select an file name...",
                          FileDialog.SAVE);

        dialog.setVisible (true);

        if (dialog.getFile () == null || dialog.getFile ().length () == 0) {
          new MessageDialog (this, "no file name given for save");
          return;
        } else {
          final String file_name = dialog.getDirectory () +
            java.io.File.separator + dialog.getFile ();

          if (new File (file_name).exists ()) {
            final YesNoDialog yes_no_dialog =
              new YesNoDialog (this, "this file exists: " + file_name +
                               " overwrite it?");

            if (yes_no_dialog.getResult ()) {
              // yes - continue
            } else {
              // no
              return;
            }
          }

          final MessageDialog message;

          message = new MessageDialog (this,
                                       "saving to " + file_name + " ...",
                                       false);

          try {
            entry.save (file_name, DocumentEntryFactory.ANY_FORMAT, false);
          } catch (EntryInformationException e) {
            final YesNoDialog yes_no_dialog =
              new YesNoDialog (this, "destination format can't handle all " +
                               "keys/qualifiers continue?");
            if (yes_no_dialog.getResult ()) {
              try {
                entry.save (file_name, DocumentEntryFactory.ANY_FORMAT, true);
              } catch (EntryInformationException e2) {
                throw new Error ("internal error - unexpected exception: "+ e);
              }
            } else {
              return;
            }
          } finally {
            message.dispose ();
          }
        }
      } else {
        final MessageDialog message =
          new MessageDialog (this,
                             "saving to " + entry.getName () + " ...",
                             false);

        entry.save (DocumentEntryFactory.ANY_FORMAT);

        message.dispose ();
      }
    } catch (ReadOnlyException e) {
      new MessageDialog (this, "this entry is read only");
      return;
    } catch (IOException e) {
      new MessageDialog (this, "error while saving: " + e);
      return;
    } catch (EntryInformationException e) {
      new MessageDialog (this, "error while saving: " + e);
      return;
    }
  }


  /**
   *  Make a new File menu replacing the current one (if any).
   **/
  private void makeFileMenu () {
    file_menu.removeAll ();

    for (int top_or_bottom = 0 ; top_or_bottom <= 1 ; ++ top_or_bottom) {
      // top_or_bottom == 0 means top (subject) otherwise bottom (query)

      final String top_or_bottom_string;
      final EntryGroup entry_group;

      if (top_or_bottom == 0) {
        top_or_bottom_string = "Top";
        entry_group = getSubjectEntryGroup ();
      } else {
        top_or_bottom_string = "Bottom";
        entry_group = getQueryEntryGroup ();
      }

      final MenuItem save_menu =
        new MenuItem ("Save All (" + top_or_bottom_string + ")");

      save_menu.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent event) {

        }
      });

      for (int source_index = 0 ;
           source_index < entry_sources.size () ;
           ++source_index) {
        final EntrySource this_source =
          entry_sources.elementAt (source_index);

        String entry_source_name = this_source.getSourceName ();

        String menu_item_name = null;

        if (entry_source_name.equals ("Filesystem")) {
          menu_item_name =
            "Read An Entry Into " + top_or_bottom_string + " ...";
        } else {
          menu_item_name =
            "Read " + entry_source_name + " Entry Into " +
            top_or_bottom_string + " ...";
        }

        final MenuItem read_entry = new MenuItem (menu_item_name);

        read_entry.addActionListener (new ActionListener () {
          public void actionPerformed (ActionEvent event) {
            try {
              final Entry new_entry =
                this_source.getEntry (entry_group.getBases ());

              if (new_entry != null) {
                entry_group.add (new_entry);
              }
            } catch (OutOfRangeException e) {
              new MessageDialog (Comparator.this,
                                 "read failed: one of the features in the" +
                                 " entry has an out of range location: " +
                                 e.getMessage ());
              return;
            } catch (IOException e) {
              new MessageDialog (Comparator.this,
                                 "read failed due to an IO error: " +
                                 e.getMessage ());
              return;
            }
          }
        });

        file_menu.add (read_entry);
      }


      if (entry_group == null || entry_group.size () == 0) {
//        read_features_menu.add (new MenuItem ("(No Entries Currently)"));
//        save_entry_menu.add (new MenuItem ("(No Entries Currently)"));
//        save_as_menu.add (new MenuItem ("(No Entries Currently)"));
      } else {

        final Menu save_entry_menu =
          new Menu ("Save " + top_or_bottom_string + " Entry");

        for (int i = 0 ; i < entry_group.size () ; ++i) {
          final Entry this_entry = entry_group.elementAt (i);

          String entry_name = this_entry.getName ();

          if (entry_name == null) {
            entry_name = "no name";
          }

          final ActionListener save_entry_listener =
            new ActionListener () {
              public void actionPerformed (final ActionEvent event) {
                Comparator.this.saveEntry (this_entry);
              }
            };

          final MenuItem save_entry_item = new MenuItem (entry_name);

          save_entry_item.addActionListener (save_entry_listener);

          save_entry_menu.add (save_entry_item);
        }

        file_menu.add (save_entry_menu);
      }

      file_menu.addSeparator ();
    }

    final MenuItem edit_subject = new MenuItem ("Edit Subject in Artemis");
    edit_subject.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        final EntryEdit entry_edit = new EntryEdit (getSubjectEntryGroup ());

        entry_edit.setVisible (true);
      }
    });

    file_menu.add (edit_subject);


    final MenuItem edit_query = new MenuItem ("Edit Query in Artemis");
    edit_query.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        final EntryEdit entry_edit = new EntryEdit (getQueryEntryGroup ());

        entry_edit.setVisible (true);
      }
    });

    file_menu.add (edit_query);


    file_menu.addSeparator ();

    final MenuItem close = new MenuItem ("Close");
    close.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        Comparator.this.dispose ();
      }
    });

    file_menu.add (close);
  }

  /**
   *  A vector of those objects listening for selection change events.
   **/
  final private Vector goto_listener_list = new Vector ();

  /**
   *  One of the two Entry objects that we are comparing.
   **/
  final private EntryGroup subject_entry_group;

  /**
   *  One of the two Entry objects that we are comparing.
   **/
  final private EntryGroup query_entry_group;

  /**
   *  The comparison data that will be displayed in the AlignmentViewer
   *  component.
   **/
  final private ComparisonData comparison_data;

  final private FeatureDisplay subject_feature_display;
  final private FeatureDisplay query_feature_display;
  final private AlignmentViewer alignment_viewer;

  private final MenuBar menu_bar = new MenuBar ();
  private final Menu file_menu   = new Menu ("File");
  private final Menu subject_graphs_menu = new Menu ("Graphs(top)");
  private final Menu query_graphs_menu  = new Menu ("Graphs(bottom)");

  final private BasePlotGroup subject_plots;
  final private BasePlotGroup query_plots;

  final private Panel button_panel;

  /**
   *  The Selection used by the subject FeatureDisplay.
   **/
  private Selection subject_selection;

  /**
   *  The Selection used by the query FeatureDisplay.
   **/
  private Selection query_selection;

  /**
   *  The GotoEventSource used by the subject FeatureDisplay.
   **/
  private GotoEventSource subject_goto_event_source;

  /**
   *  The GotoEventSource used by the query FeatureDisplay.
   **/
  private GotoEventSource query_goto_event_source;

  /**
   *  The EntrySourceVector reference that is created in the constructor.
   **/
  private EntrySourceVector entry_sources;
}
