/* MultiComparator.java
 *
 * created: Tue Sep 11 2001
 *
 * This file is part of Artemis
 *
 * Copyright (C) 2001  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/pathsoft/repository/powmap/diana/components/MultiComparator.java,v 1.1 2001/09/13 11:37:09 kmr Exp $
 */

package diana.components;

import AppGlobal;

import diana.*;
import diana.plot.*;
import diana.sequence.*;

import uk.ac.sanger.pathogens.OutOfRangeException;
import uk.ac.sanger.pathogens.embl.EntryInformationException;
import uk.ac.sanger.pathogens.embl.Range;
import uk.ac.sanger.pathogens.embl.DocumentEntryFactory;
import uk.ac.sanger.pathogens.ReadOnlyException;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Vector;
import java.util.EventObject;

/**
 *  This Frame component contains an arbitrary number of AlignmentViewer
 *  components and FeatureDisplay components along with ComparatorGlue objects
 *  to keep them synchronized.
 *
 *  @author Kim Rutherford <kmr@sanger.ac.uk>
 *  @version $Id: MultiComparator.java,v 1.1 2001/09/13 11:37:09 kmr Exp $
 **/

public class MultiComparator extends Frame
    implements FeatureDisplayOwner {
  /**
   *  Initialise entry_group_array and comparison_data_array and create all
   *  the FeatureDisplay and AlignmentViewer components.
   **/
  public MultiComparator (final EntryGroup [] entry_group_array,
                          final ComparisonData [] comparison_data_array) {
    if (entry_group_array.length != comparison_data_array.length + 1) {
      throw new Error ("internal error - " +
                       "MultiComparator got illegal arguments");
    }

    this.entry_group_array = entry_group_array;
    this.comparison_data_array = comparison_data_array;

    final StringBuffer title_buffer = new StringBuffer ();
 
    title_buffer.append ("ACT: ");

    for (int i = 0 ; i < entry_group_array.length - 1 ; ++i) {
      final String sequence_name =
        entry_group_array[i].getDefaultEntry ().getName ();
      title_buffer.append (sequence_name).append (" vs ");
    }

    final EntryGroup last_entry_group =
      entry_group_array[entry_group_array.length - 1];

    title_buffer.append (last_entry_group.getDefaultEntry ().getName ());

    setTitle (title_buffer.toString ());

    makeSelectionArray ();
    makeGotoEventSourceArray ();

    feature_display_array =
      new FeatureDisplay [entry_group_array.length];
    base_plot_group_array =
      new BasePlotGroup [entry_group_array.length];
    alignment_viewer_array =
      new AlignmentViewer [comparison_data_array.length];
    comparator_glue_array =
      new ComparatorGlue [comparison_data_array.length];

    for (int i = 0 ; i < getEntryGroupArray ().length ; ++i) {
      feature_display_array[i] =
        new FeatureDisplay (getEntryGroupArray ()[i],
                            getSelectionArray ()[i],
                            getGotoEventSourceArray ()[i],
                            this,
                            // put scrollbar at the bottom normally
                            !(i == getEntryGroupArray ().length - 1 && 
                              getEntryGroupArray ().length == 2));

      feature_display_array[i].setShowLabels (false);
      if (getEntryGroupArray ().length > 2) {
        feature_display_array[i].setShowForwardFrameLines (false);
        feature_display_array[i].setShowReverseFrameLines (false);
      }

      final BasePlotGroup base_plot_group =
        new BasePlotGroup (getEntryGroupArray ()[i], this,
                           getSelectionArray ()[i],
                           getGotoEventSourceArray ()[i]);

      feature_display_array[i].addDisplayAdjustmentListener (base_plot_group);
      
      base_plot_group_array[i] = base_plot_group;
    }

    for (int i = 0 ; i < getComparisonDataArray ().length ; ++i) {
      alignment_viewer_array[i] =
        new AlignmentViewer (getFeatureDisplayArray ()[i] ,
                             getFeatureDisplayArray ()[i + 1],
                             getComparisonDataArray ()[i]);

      comparator_glue_array[i] =
        new ComparatorGlue (this,
                            getFeatureDisplayArray ()[i],
                            getFeatureDisplayArray ()[i + 1],
                            getAlignmentViewerArray ()[i]);
    }

    final Font default_font = getDefaultFont ();

    setFont (default_font);

    makeMenus ();

    GridBagLayout gridbag = new GridBagLayout();
    setLayout (gridbag);

    GridBagConstraints c = new GridBagConstraints();

    c.gridwidth  = GridBagConstraints.REMAINDER;
    c.fill       = GridBagConstraints.HORIZONTAL;
    c.anchor     = GridBagConstraints.NORTH;
    c.gridheight = 1;
    c.weightx    = 1;
    c.weighty    = 0;

    for (int i = 0 ; i < getFeatureDisplayArray ().length ; ++i) {
      if (!(i == getEntryGroupArray ().length - 1 && 
            getEntryGroupArray ().length == 2)) {
        // put graph above the sequence in this case
        c.weighty = 0;
        gridbag.setConstraints (base_plot_group_array[i], c);
        add (base_plot_group_array[i]);
      }

      c.weighty = 0;
      gridbag.setConstraints (feature_display_array[i], c);
      add (feature_display_array[i]);

      if (i == getEntryGroupArray ().length - 1 && 
            getEntryGroupArray ().length == 2) {
        // put graph below the sequence in this case
        c.weighty = 0;
        gridbag.setConstraints (base_plot_group_array[i], c);
        add (base_plot_group_array[i]);
      }

      if (i < getAlignmentViewerArray ().length) {
        c.fill = GridBagConstraints.BOTH;
        c.weighty = 1;
        gridbag.setConstraints (getAlignmentViewerArray ()[i], c);
        add (getAlignmentViewerArray ()[i]);
      }
    }

    addWindowListener (new WindowAdapter () {
      public void windowClosing (WindowEvent event) {
        MultiComparator.this.dispose ();
      }
    });

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();

    packme ();

    int screen_height = screen.height;
    int screen_width = screen.width;

    if (getAlignmentViewerArray ().length > 0) {
      if (screen_width <= 900 || screen_height <= 700) {
        setSize (screen_width * 9 / 10, screen_height * 9 / 10);
      } else {
        setSize (900, 700);
      }
    }

    setLocation (new Point ((screen.width - getSize ().width) / 2,
                            (screen.height - getSize ().height) / 2));
  }

  /**
   *  Make and add the menus for this component.
   **/
  private void makeMenus () {
    final Font default_font = getDefaultFont ();

    setMenuBar (menu_bar);

    //  makeFileMenu ();
    
    //  menu_bar.add (file_menu);

    for (int i = 0 ; i < getFeatureDisplayArray ().length ; ++i) {
      final BasePlotGroup base_plot_group = getBasePlotGroupArray ()[i];

      final BaseAlgorithm [] group_algorithms =
        base_plot_group.getPlotAlgorithms ();

      final Menu this_graphs_menu = new Menu ("Graphs(" + (i + 1) + ")");

      for (int algorithm_index = 0 ;
           algorithm_index < group_algorithms.length ;
           ++algorithm_index) {
        final BaseAlgorithm this_algorithm =
          group_algorithms[algorithm_index];

        addDisplayMenuAlgorithm (base_plot_group, this_algorithm,
                                 this_graphs_menu);
      }

      menu_bar.add (this_graphs_menu);
    }
    

    final Menu display_menu = new Menu ("Display");


    final MenuItem hide_on_frame_lines_item =
      new MenuItem ("Hide Frame Lines");
    hide_on_frame_lines_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        for (int i = 0 ; i < getFeatureDisplayArray ().length ; ++i) {
          getFeatureDisplayArray ()[i].setShowForwardFrameLines (false);
          getFeatureDisplayArray ()[i].setShowReverseFrameLines (false);
        }
      }
    });

    display_menu.add (hide_on_frame_lines_item);


    final MenuItem show_on_frame_lines_item =
      new MenuItem ("Show Frame Lines");
    show_on_frame_lines_item.addActionListener (new ActionListener () {
      public void actionPerformed (ActionEvent event) {
        for (int i = 0 ; i < getFeatureDisplayArray ().length ; ++i) {
          getFeatureDisplayArray ()[i].setShowForwardFrameLines (true);
          getFeatureDisplayArray ()[i].setShowReverseFrameLines (true);
        }
      }
    });

    display_menu.add (show_on_frame_lines_item);


    menu_bar.add (display_menu);
  }

  /**
   *  This method will call pack () and then move the Frame to the centre of
   *  the screen.  (Implementation of the FeatureDisplayOwner interface).
   **/
  public void packme () {
    pack ();

    final Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();

    final int x_position = (screen.width - getSize ().width) / 2;
    int y_position = (screen.height - getSize ().height) / 2;

    if (y_position < 10) {
      y_position = 10;
    }

    setLocation (new Point (x_position, y_position));
  }

  /**
   *  Make one Selection object each EntryGroup.
   **/
  private void makeSelectionArray () {
    selection_array = new Selection [getEntryGroupArray ().length];
    for (int i = 0 ; i < selection_array.length ; ++i) {
      selection_array[i] = new Selection (null);
    }
  }

  /**
   *  Make subject_goto_event_source and query_goto_event_source and wire them
   *  togeather so that goto events do the right thing.
   **/
  private void makeGotoEventSourceArray () {
    goto_event_source_array =
      new GotoEventSource [getEntryGroupArray ().length];

    for (int i = 0 ; i < goto_event_source_array.length ; ++i) {
      goto_event_source_array[i] =
        new SimpleGotoEventSource (getEntryGroupArray ()[i]) {
          /**
           *  Send the given event to all the GotoListeners.
           **/
          public void sendGotoEvent (final GotoEvent goto_event) {
            // temporarily disable moving so that the other FeatureDisplay
            // doesn't start moving (because of the listeners set up in
            // addDisplayListeners ()
            super.sendGotoEvent (goto_event);
          }
        };
    }
  }

  /**
   *  Add a menu item for the given algorithm to the given menu.
   **/
  private void addDisplayMenuAlgorithm (final BasePlotGroup base_plot_group,
                                        final BaseAlgorithm algorithm,
                                        final Menu menu) {
    final CheckboxMenuItem new_item =
      new CheckboxMenuItem (algorithm.getAlgorithmName ());

    new_item.setState (base_plot_group.basePlotIsVisible (algorithm));

    new_item.addItemListener (new ItemListener () {
      public void itemStateChanged(ItemEvent event) {
        base_plot_group.setVisibleByAlgorithm (algorithm,
                                               new_item.getState ());
        packme ();
      }
    });

    menu.add (new_item);
  }

  /**
   *  Save the given entry, prompting for a file name if necessary.
   **/
  private void saveEntry (final Entry entry) {
    try {
      if (entry.getName () == null) {
        // prompt for a name
        final FileDialog dialog =
          new FileDialog (this,
                          "Select an file name...",
                          FileDialog.SAVE);

        dialog.setVisible (true);

        if (dialog.getFile () == null || dialog.getFile ().length () == 0) {
          new MessageDialog (this, "no file name given for save");
          return;
        } else {
          final String file_name = dialog.getDirectory () +
            java.io.File.separator + dialog.getFile ();

          if (new File (file_name).exists ()) {
            final YesNoDialog yes_no_dialog =
              new YesNoDialog (this, "this file exists: " + file_name +
                               " overwrite it?");

            if (yes_no_dialog.getResult ()) {
              // yes - continue
            } else {
              // no
              return;
            }
          }

          final MessageDialog message;

          message = new MessageDialog (this,
                                       "saving to " + file_name + " ...",
                                       false);

          try {
            entry.save (file_name, DocumentEntryFactory.ANY_FORMAT, false);
          } catch (EntryInformationException e) {
            final YesNoDialog yes_no_dialog =
              new YesNoDialog (this, "destination format can't handle all " +
                               "keys/qualifiers continue?");
            if (yes_no_dialog.getResult ()) {
              try {
                entry.save (file_name, DocumentEntryFactory.ANY_FORMAT, true);
              } catch (EntryInformationException e2) {
                throw new Error ("internal error - unexpected exception: "+ e);
              }
            } else {
              return;
            }
          } finally {
            message.dispose ();
          }
        }
      } else {
        final MessageDialog message =
          new MessageDialog (this,
                             "saving to " + entry.getName () + " ...",
                             false);

        entry.save (DocumentEntryFactory.ANY_FORMAT);

        message.dispose ();
      }
    } catch (ReadOnlyException e) {
      new MessageDialog (this, "this entry is read only");
      return;
    } catch (IOException e) {
      new MessageDialog (this, "error while saving: " + e);
      return;
    } catch (EntryInformationException e) {
      new MessageDialog (this, "error while saving: " + e);
      return;
    }
  }


  /**
   *  Make a new File menu replacing the current one (if any).
   **/
  private void makeFileMenu () {
//      file_menu.removeAll ();

//      for (int top_or_bottom = 0 ; top_or_bottom <= 1 ; ++ top_or_bottom) {
//        // top_or_bottom == 0 means top (subject) otherwise bottom (query)

//        final String top_or_bottom_string;
//        final EntryGroup entry_group;

//        if (top_or_bottom == 0) {
//          top_or_bottom_string = "Top";
//          entry_group = getSubjectEntryGroup ();
//        } else {
//          top_or_bottom_string = "Bottom";
//          entry_group = getQueryEntryGroup ();
//        }

//        final MenuItem save_menu =
//          new MenuItem ("Save All (" + top_or_bottom_string + ")");

//        save_menu.addActionListener (new ActionListener () {
//          public void actionPerformed (ActionEvent event) {

//          }
//        });

//        for (int source_index = 0 ;
//             source_index < entry_sources.size () ;
//             ++source_index) {
//          final EntrySource this_source =
//            entry_sources.elementAt (source_index);

//          String entry_source_name = this_source.getSourceName ();

//          String menu_item_name = null;

//          if (entry_source_name.equals ("Filesystem")) {
//            menu_item_name =
//              "Read An Entry Into " + top_or_bottom_string + " ...";
//          } else {
//            menu_item_name =
//              "Read " + entry_source_name + " Entry Into " +
//              top_or_bottom_string + " ...";
//          }

//          final MenuItem read_entry = new MenuItem (menu_item_name);

//          read_entry.addActionListener (new ActionListener () {
//            public void actionPerformed (ActionEvent event) {
//              try {
//                final Entry new_entry =
//                  this_source.getEntry (entry_group.getBases ());

//                if (new_entry != null) {
//                  entry_group.add (new_entry);
//                }
//              } catch (OutOfRangeException e) {
//                new MessageDialog (Comparator.this,
//                                   "read failed: one of the features in the" +
//                                   " entry has an out of range location: " +
//                                   e.getMessage ());
//                return;
//              } catch (IOException e) {
//                new MessageDialog (Comparator.this,
//                                   "read failed due to an IO error: " +
//                                   e.getMessage ());
//                return;
//              }
//            }
//          });

//          file_menu.add (read_entry);
//        }


//        if (entry_group == null || entry_group.size () == 0) {
//          //        read_features_menu.add (new MenuItem ("(No Entries Currently)"));
//          //        save_entry_menu.add (new MenuItem ("(No Entries Currently)"));
//          //        save_as_menu.add (new MenuItem ("(No Entries Currently)"));
//        } else {

//          final Menu save_entry_menu =
//            new Menu ("Save " + top_or_bottom_string + " Entry");

//          for (int i = 0 ; i < entry_group.size () ; ++i) {
//            final Entry this_entry = entry_group.elementAt (i);

//            String entry_name = this_entry.getName ();

//            if (entry_name == null) {
//              entry_name = "no name";
//            }

//            final ActionListener save_entry_listener =
//              new ActionListener () {
//                public void actionPerformed (final ActionEvent event) {
//                  Comparator.this.saveEntry (this_entry);
//                }
//              };

//            final MenuItem save_entry_item = new MenuItem (entry_name);

//            save_entry_item.addActionListener (save_entry_listener);

//            save_entry_menu.add (save_entry_item);
//          }

//          file_menu.add (save_entry_menu);
//        }

//        file_menu.addSeparator ();
//      }

//      final MenuItem edit_subject = new MenuItem ("Edit Subject in Artemis");
//      edit_subject.addActionListener (new ActionListener () {
//        public void actionPerformed (ActionEvent event) {
//          final EntryEdit entry_edit = new EntryEdit (getSubjectEntryGroup ());

//          entry_edit.setVisible (true);
//        }
//      });

//      file_menu.add (edit_subject);


//      final MenuItem edit_query = new MenuItem ("Edit Query in Artemis");
//      edit_query.addActionListener (new ActionListener () {
//        public void actionPerformed (ActionEvent event) {
//          final EntryEdit entry_edit = new EntryEdit (getQueryEntryGroup ());

//          entry_edit.setVisible (true);
//        }
//      });

//      file_menu.add (edit_query);


//      file_menu.addSeparator ();

//      final MenuItem close = new MenuItem ("Close");
//      close.addActionListener (new ActionListener () {
//        public void actionPerformed (ActionEvent event) {
//          Comparator.this.dispose ();
//        }
//      });

//      file_menu.add (close);
  }

  /**
   *  Return the current default font (from Diana.options).
   **/
  private Font getDefaultFont () {
    return Options.getOptions ().getFont ();
  }

  /**
   *  Return the EntryGroup objects that were passed to the constructor.
   **/
  private EntryGroup [] getEntryGroupArray () {
    return entry_group_array;
  }

  /**
   *  Return the ComparisonData objects that were passed to the constructor.
   **/
  private ComparisonData [] getComparisonDataArray () {
    return comparison_data_array;
  }

  /**
   *  Return the AlignmentViewer objects that were created in the constructor.
   **/
  private AlignmentViewer [] getAlignmentViewerArray () {
    return alignment_viewer_array;
  }

  /**
   *  Return the FeatureDisplay objects that were created in the constructor.
   **/
  private FeatureDisplay [] getFeatureDisplayArray () {
    return feature_display_array;
  }

  /**
   *  Return the Selection objects that were created in the constructor.
   **/
  private Selection [] getSelectionArray () {
    return selection_array;
  }

  /**
   *  Return the GotoEventSource objects that were created in the constructor.
   **/
  private GotoEventSource [] getGotoEventSourceArray () {
    return goto_event_source_array;
  }

  /**
   *  Return the BasePlotGroup objects that were created in the constructor.
   **/
  private BasePlotGroup [] getBasePlotGroupArray () {
    return base_plot_group_array;
  }

  /**
   *  The menu_bar is created in makeMenus ().
   **/
  private final MenuBar menu_bar = new MenuBar ();

  /**
   *  An array of EntryGroup objects to display (set by the constructor).  The
   *  array will contain at least one EntryGroup and will be exactly one
   *  element longer than comparison_data_array.
   **/
  private EntryGroup [] entry_group_array = null;

  /**
   *  An array of ComparisonData objects to display (set by the constructor).
   *  The array will be exactly one element shorter than entry_group_array.
   **/
  private ComparisonData [] comparison_data_array = null;

  /**
   *  An array of Selection objects - one per EntryGroup.  Created by
   *  makeSelectionArray()
   **/
  private Selection [] selection_array = null;

  /**
   *  An array of GotoEventSource objects - one per EntryGroup.  Created by
   *  makeGotoEventSourceArray()
   **/
  private GotoEventSource [] goto_event_source_array = null;

  /**
   *  An array of FeatureDisplay objects - one per EntryGroup.  Created in the
   *  constructor.
   **/
  private FeatureDisplay [] feature_display_array = null;

  /**
   *  An array of AlignmentViewer objects - one per ComparisonData.  Created
   *  in the constructor.
   **/
  private AlignmentViewer [] alignment_viewer_array = null;

  /**
   *  An array of AlignmentViewer objects - one per ComparisonData.  Created
   *  in the constructor.
   **/
  private ComparatorGlue [] comparator_glue_array = null;

  /**
   *  An array of BasePlotGroup objects - one per FeatureDisplay/EntryGroup.
   *  Created in the constructor.
   **/
  private BasePlotGroup [] base_plot_group_array = null;
}
